import joblib
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import pickle
from Crypto.Cipher import AES
from Crypto.Hash import SHA256
from sklearn.mixture import GaussianMixture
from .utilitymod import TRC
from .utilitymod import FiberLines
from .utilitymod import Pixels
from .utilitymod import Lines
from .utilitymod import Particle
from .utilitymod import PMA
from .utilitymod import line
from typing import NamedTuple
from scipy import signal
from tqdm.notebook import trange


class PennateAngle(object):
    def __init__(self, filepath):
        self.filepath = filepath
        self.gsv = None
        self.gsvave = None
        self.radonn = None
        self.G1 = None
        self.G2 = None
        self.ratio = 0.2
        self.lowerlimit = 30
        self.upperlimit = 150
        self.resGMM = None
        self.resGMM_ = None
        self.set_data_simple()
        self.set_data()

    def read_data(self):
        with open(self.filepath, "rb") as f:
            a =  pickle.load(f)
        return a

    def set_ratio(self, ratio):
        self.ratio = ratio
        print("The ratio has been set successfully.")

    def set_limit(self, lowerlimit, upperlimit):
        self.lowerlimit = lowerlimit
        self.upperlimit = upperlimit
        print("The limit values have been set successfully.")

    def create_aes(self, password, iv):
        sha = SHA256.new()
        sha.update(password.encode())
        key = sha.digest()
        return AES.new(key, AES.MODE_CFB, iv)

    def decrypt(self, data, shape, dtype):
        iv, cipher = data[:AES.block_size], data[AES.block_size:]
        res = self.create_aes("test_passward", iv).decrypt(cipher)
        return np.frombuffer(res, dtype).reshape(-1, shape[0], shape[1])

    def set_data_simple(self):
        data = self.read_data()
        self.G1 = self.decrypt(data["G1"], data["G1_shape"], data["G1_dtype"])
        self.G2 = self.decrypt(data["G2"], data["G2_shape"], data["G2_dtype"])
        self.fcount = self.G1.shape[0]
        self.shape2 = [self.G1.shape[1], self.G1.shape[2]]
        self.axlim2 = [(self.shape2[0], 0), (0, self.shape2[1])]
        del data

    def set_data(self):
        data = self.read_data()
        self.gsv = self.decrypt(data["gsv"], data["gsv_shape"], data["gsv_dtype"])
        self.gsvave = self.decrypt(data["gsvave"], data["gsvave_shape"], data["gsvave_dtype"])
        self.radonn = self.decrypt(data["radonn"], data["radonn_shape"], data["radonn_dtype"])
        self.shape1 = [self.gsv.shape[1], self.gsv.shape[2]]
        self.axlim1 = [(self.shape1[0], 0), (0, self.shape1[1])]
        del data



    def getPixels(self, frame):
        val = sorted(self.G1[frame].ravel())[-int(self.G1[frame].size * self.ratio * 0.01)::]
        y, x = np.where(self.G1[frame] >= val[0])
        return x, y

    def getPixelsGMM(self, *frame, mav=5, f1=42, f2=100):
        """
        frame : int
        mav : int
        f1 : us frequency
        f2 : vicon frequency
        """
        if frame == ():
            if self.resGMM == None:
                sa, pm, da = [], [], []
                for f in trange(self.fcount):
                    theta, rho = self.getPixels(f)
                    _upperlimit= list(np.where(theta > self.upperlimit)[0])
                    _lowerlimit = list(np.where(theta < self.lowerlimit)[0])
                    _upperlimit.extend(_lowerlimit)
                    theta = np.delete(theta, _upperlimit)
                    rho = np.delete(rho, _upperlimit)
                    cdata = np.array([theta, rho]).T
                    res = GaussianMixture(n_components=3).fit_predict(cdata)
                    sign = np.array([cdata[res==i].mean(axis=0)[1] for i in range(3)])
                    res = [TRC(cdata[res==np.argsort(sign)[i]]) for i in range(3)]
                    sa.append(res[0].theta_)
                    pm.append(res[1].theta_)
                    da.append(res[2].theta_)
                sa, da, pm = -np.array(sa), -np.array(da), -np.array(pm)
                pma = pm - da
                pma_ = np.convolve(pma, np.ones(mav)/mav, mode="same")
                lcm = lcm = np.lcm(f1, f2)
                pm_f3 = signal.resample(pma_, pm.size *(lcm // f1))
                pm_100hz = signal.resample(pm_f3, pm_f3.size//(lcm//f2))
                result = PMA(line(da), line(pm), line(sa), line(pma), line(pma_), line(pm_100hz))
                #result = PMA(line( -np.array(da)), line(-np.array(pm)),line(-np.array(sa)))
                self.resGMM = result
                return result
            else:
                return self.resGMM
        else:
            frame = frame[0]
            theta, rho = self.getPixels(frame)
            _upperlimit= list(np.where(theta > self.upperlimit)[0])
            _lowerlimit = list(np.where(theta < self.lowerlimit)[0])
            _upperlimit.extend(_lowerlimit)
            theta = np.delete(theta, _upperlimit)
            rho = np.delete(rho, _upperlimit)
            cdata = np.array([theta, rho]).T
            res = GaussianMixture(n_components=3).fit_predict(cdata)
            sign = np.array([cdata[res==i].mean(axis=0)[1] for i in range(3)])
            res = [TRC(cdata[res==np.argsort(sign)[i]]) for i in range(3)]
            return Pixels(res[2], res[1], res[0])
    

    #def _getLine(self, frame, fibertype):
    #    val = self.getPixelsGMM(frame)
    #    val = getattr(val, fibertype)
    #    Yshape, Xshape = self.gsv[0].shape

    #    x = np.arange(0, Xshape, 1)
    #    positive_, negative_ = [], []
    #    for i, j in zip(val.theta, val.rho):
    #        tilt = np.radians(90 - i)
    #        if np.radians(90 - i) >= 0:
    #            positive_.append(np.tan(tilt)*(x - Xshape/2) + Yshape/2 - (j - self.G1[0].shape[0]/2))
    #        else:
    #            negative_.append(np.tan(tilt)*(x - Xshape/2) + Yshape/2 - (j - self.G1[0].shape[0]/2))
    #    return Lines(positive_, negative_)

    #def _getLine(self, frame, val):
    #    Yshape, Xshape = self.gsv[0].shape
    #    x = np.arange(0, Xshape, 1)
    #    positive_, negative_ = [], []
    #    for i, j in zip(val.theta, val.rho):
    #        tilt = np.radians(90 - i)
    #        if np.radians(90 - i) >= 0:
    #            positive_.append(np.tan(tilt)*(x - Xshape/2) + Yshape/2 - (j - self.G1[0].shape[0]/2))
    #        else:
    #            negative_.append(np.tan(tilt)*(x - Xshape/2) + Yshape/2 - (j - self.G1[0].shape[0]/2))
    #    return Lines(positive_, negative_)

    def _getLine(self, frame, val):
        Yshape, Xshape = self.gsv[0].shape
        x = np.arange(0, Xshape, 1)
        res = []
        for i, j in zip(val.theta, val.rho):
            tilt = np.radians(90 - i)
            res.append(np.tan(tilt)*(x - Xshape/2) + Yshape/2 - (j - self.G1[0].shape[0]/2))
        return Particle(np.array(res))

    #def _getAverageLine(self, frame, fibertype):
    #    val = self.getPixelsGMM(frame)
    #    val = getattr(val, fibertype)
    #    Yshape, Xshape = self.gsv[0].shape
    #    x = np.arange(0, Xshape, 1)
    #    return np.tan(np.radians(val.theta_))*(x - Xshape/2) + Yshape/2 - (val.rho_ - self.G1[0].shape[0]/2)

    def _getAverageLine(self, frame, val):
        Yshape, Xshape = self.gsv[0].shape
        x = np.arange(0, Xshape, 1)
        res = np.tan(np.radians(val.theta_))*(x - Xshape/2) + Yshape/2 - (val.rho_ - self.G1[0].shape[0]/2)
        return Particle(res)

    #def getFiberLines(self, frame):
    #    a = [self._getLine(frame, i) for i in ["sa", "pm", "da"]]
    #    b = [self._getAverageLine(frame, i) for i in ["sa", "pm", "da"]]
    #    a.extend(b)
    #    return FiberLines(*a)

    def getFiberLines(self, frame):
        temp = self.getPixelsGMM(frame)
        a = [self._getLine(frame, getattr(temp, i)) for i in ["sa", "pm", "da"]]
        b = [self._getAverageLine(frame, getattr(temp, i)) for i in ["sa", "pm", "da"]]
        a.extend(b)
        return FiberLines(*a)


    def weight_ave(self, data):
        res = []
        for i in data:
            a, b = np.unique(i, return_counts=True)
            res.append(np.average(a, weights=b))
        return np.array(res)

    def calc_outlier(self, data):
        Q1 = np.percentile(data, 25)
        Q3 = np.percentile(data, 75)
        IQR = Q3 - Q1
        LOL = Q1 - 1.5 * IQR
        HOL = Q3 + 1.5 * IQR
        return np.where((data < LOL) | (data > HOL))[0], LOL, HOL

    def analyze_result(self, ftype):
        namelist = {"pm":"Pennate muscle",
                    "sa":"sup aponeurosis",
                    "da":"dep aponeurosis"
                }
        if self.resGMM_ == None:
            res = []
            for i in trange(self.fcount):
                res.append(self.getPixelsGMM(i))
            self.resGMM_ = res
        else:
            res = self.resGMM_

        count, theta, rho = [], [], []
        for i in res:
            count.append(getattr(getattr(i, ftype), "theta").size)
            theta.append(getattr(getattr(i, ftype), "theta"))
            rho.append(getattr(getattr(i, ftype), "rho"))

        theta_, rho_, count_ = self.weight_ave(theta), self.weight_ave(rho), self.weight_ave(count)
        theta_ol, theta_lol, theta_hol  = self.calc_outlier(theta_)
        rho_ol, rho_lol, rho_hol= self.calc_outlier(rho_)
        count_ol, count_lol, count_hol = self.calc_outlier(count_)

        outlier_rate = np.round(np.array([theta_ol.size/theta_.size, rho_ol.size/rho_.size, count_ol.size/count_.size])*100,2)
        lol = np.array([theta_lol, rho_lol, count_lol])
        hol = np.array([theta_hol, rho_hol, count_hol])
        df = pd.DataFrame({
            "outlier rate(%)":outlier_rate,
            "Lower limit":lol,
            "Higher limit":hol},
            index=[r"angle ($\theta$)", "distance (pixels)", "count"])

        print("="*40)
        print(f"Result of {namelist[ftype]} data analysis")
        print("="*40)
        display(df)
        print("")
        print("Outlier frame")
        print("="*40)
        print("theta")
        print(theta_ol)
        print("rho")
        print(rho_ol)
        print("count")
        print(count_ol)
        
        if all(df["outlier rate(%)"].values < 5):
            print("As a result of the analysis, the measurement accuracy is sufficient.")
        else:
            print("The measurement accuracy is low. Remeasurement is recommended.")

        pix = self.fcount
        itheta = np.zeros((int(pix/20),pix))
        irho = np.zeros((int(pix/20),pix))
        icount = np.zeros((int(pix/20),pix))
        #x = np.arange(pix)
        for img, d in zip([itheta, irho, icount],[theta_, rho_, count_]):
            for i in range(pix):
                for j in range(int(pix/20)):
                    img[j][i] = d[i]
        plt.figure(figsize=(12,8))
        plt.subplot(3,1,1)
        plt.title(r"angle [$\theta$]", fontsize=15)
        plt.imshow(itheta, cmap="Blues")
        for i in theta_ol:
            plt.axvline(i, c="r")
        plt.colorbar(orientation="horizontal")
        plt.subplot(3,1,2)
        plt.title("distance [pixel]", fontsize=15)
        plt.imshow(irho, cmap="Blues")
        for i in rho_ol:
            plt.axvline(i, c="r")
        plt.colorbar(orientation="horizontal")
        plt.subplot(3,1,3)
        plt.title("count", fontsize=15)
        plt.imshow(icount, cmap="Blues")
        for i in count_ol:
            plt.axvline(i, c="r")
        plt.colorbar(orientation="horizontal")
        plt.tight_layout()






























    def IOI(self, frame, ratio=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6]):
        pixels = [self.getPixels(frame, i) for i in ratio]
        plt.figure(figsize=(14,14))
        plt.suptitle(f"Frame : {frame} -> Time : {frame/30:.2f}sec", fontsize=30)

        plt.subplot(2,4,1)
        plt.title("original", fontsize=15)
        plt.imshow(self.gsv[frame], cmap="gray")
        plt.grid(c="c")
        plt.axhline(136, c="r")

        plt.subplot(2,4,5)
        plt.title("substract mean", fontsize=15)
        plt.imshow(self.gsvave[frame], cmap="gray")
        plt.grid(c="c")
        plt.axhline(136, c="r")

        for num, i in enumerate([2,3,4,6,7,8]):
            plt.subplot(2,4,i)
            plt.title(f"top : {ratio[num]}% [ {pixels[num][0].size}/{self.G1[0].size} ]", fontsize=15)
            plt.imshow(self.G1[frame], cmap="gray")
            plt.scatter(pixels[num][0], pixels[num][1], facecolor="None", edgecolor="g", alpha=0.5,s=3)
            plt.grid(c="w")
            plt.axhline(193, c="r")
        plt.tight_layout(rect=[0,0,1,0.98])
        